//Este archivo contiene las funciones de insersin y bsqueda en la tabla hash

#include "variables.h"
#include <string.h>

int FuncionHashMult (unsigned long int, int, int);
int FuncionHashLineal (unsigned long int, int, int);
unsigned long int crearclave (Elemento);


//Funcn CrearTHash : Crea una nueva tabla hash
//Variables de Entrada : N - Cantidad de elementos
//Valor devuelto : La tabla hash
TablaHashT CrearTHash (int N)
{
  TablaHashT TablaHash;	//Tabla
  int indice;		//Variable indice

  TablaHash = new Elemento[N];			//Crea espacio en memoria para la tabla
  for (indice = 0; indice < N; indice++)
      TablaHash[indice].Palabra = NULL;		//Setear cada elemento de la tabla como vaco
  return (TablaHash);
}//EndCrearTHash


//Funcin InsertarHash : Inserta un elemento a la tabla hash.
//Variables de entrada : TablaHash - arreglo con tabla hash,
//			 ElementoH - Elemento a ingresar,
//			 N         - tamao de la tabla hash,
//			 tipodehash- parmetro indicador del tipo de funcin hash a usar:
//			 0 = hash multiplicativo ; 1 = hash lineal
//Variables de salida  : arreglo con tabla hash con nuevo elemento ingresado
//		       : intentos - intentos necesarios para insertar el elemento
//Valor devuelto       : 0 = xito ; 1 = Hash Overflow
int InsertarHash (TablaHashT TablaHash, Elemento ElementoH, int N, int tipodehash, int &intentos)
{
   unsigned long int clave;	//Clave del elemento a ingresar
   int indice,			//Indice de la tabla hash en el cual insertar el elemento
       prueba = 0,	    	//Numero de prueba de la clave
       insertado = 0;		//Sentinela que indica si la clave fue insertada en la tabla hash

   intentos = 0;					//Resetear la variable intentos
   clave = crearclave (ElementoH);			//Crear la clave para el elemento a insertar
   do
     {
      intentos++;					//Sumar uno a la var. intentos.
      if (tipodehash == 0)				//Mtodo multiplicativo?
        {
         indice = FuncionHashMult(clave, prueba, N); 	//Utilizar mtodo multiplicativo
        }//EndIf
       else						//Sino...
	 {
          indice = FuncionHashLineal (clave, prueba, N);//...utilizar mtodo lineal
         }//EndElse
      if (TablaHash[indice].Palabra == NULL)		//Si la posicin est libre...
        {
         TablaHash[indice].Palabra = new char[strlen(ElementoH.Palabra)]; //Crear espacio en mem.
         strcpy (TablaHash[indice].Palabra, ElementoH.Palabra);		  //...insertar elemento y...
         insertado = 1;						   	  //...marcar elemento como insertado.
	 prueba--;			//Evita que se devuelva overflow en casos que no lo hubo
	}//EndIf
      prueba++;				//Preparar prxima prueba (si la hay)
     }//EndDoWhile
   while ((insertado == 0) && (prueba < N));	//Repetir hasta insertar el elemento o detectar hash overflow
   return (prueba == N);
}//EndInsertarHash

//Funcin BuscarHash   : Busca un elemento en la tabla hash.
//Variables de entrada : TablaHash - arreglo con tabla hash,
//			 			 clave     - clave del elemento a buscar,
//			 			 N         - tamao de la tabla hash,
//			 			 tipodehash- parmetro indicador del tipo de funcin hash a usar:
//			 			 0 = hash multiplicativo ; 1 = hash lineal
//Variables de salida  : intentos - intentos necesarios para encontrar el elemento
//Valor devuelto       : indice de la tabla hash en el cual est alojado el elemento buscado (-1 si no est)
int BuscarHash (TablaHashT TablaHash, Elemento ElementoH, int N, int tipodehash, int &intentos)
{
   unsigned long int clave;	//Clave del elemento a buscar
   int indice,			//Indice de la tabla hash en el cual insertar el elemento
       prueba = 0,	    	//Numero de prueba de la clave
       encontrado = 0;		//Sentinela que indica si la clave fue encontrada en la tabla hash

   intentos = 0;					//Resetear la variable intentos.
   clave = crearclave (ElementoH);			//Crear clave para la palabra a buscar.
   do
     {
      intentos++;					//Sumar uno a la var. intentos
      if (tipodehash == 0)				//Mtodo multiplicativo?
        {
         indice = FuncionHashMult (clave, prueba, N);	//Utilizar mtodo multiplicativo
        }//EndIf
       else						//Sino...
	 {
          indice = FuncionHashLineal (clave, prueba, N);//...utilizar mtodo lineal
         }//EndElse
      if (strcmp(TablaHash[indice].Palabra,ElementoH.Palabra) == 0)	//Si encontramos el elemento...
        {
         encontrado = 1;			//...marcar elemento como encontrado.
	 prueba--;				//Evita que se devuelva "no encontrado" en casos que se encontr
	}//EndIf
      prueba++;					//Preparar prxima prueba (si la hay)
     }//EndDoWhile
   while ((encontrado == 0) && (prueba < N) && (TablaHash[indice].Palabra != NULL));	//Repetir hasta encontrar o detectar que no est

   if (encontrado == 1)			//Si el elemento se encontr...
     {
      return (indice);			//...publicitar su ndice en la tabla hash
     }//EndIf
    else				//Sino...
      {
       return (-1);			//chuminga.
      }//EndElse
}//EndBuscarHash


//Funcin DestruirTHash : Destruye una tabla hash y libera el espacio ocupado por sta
//Variables de entrada : La tabla y su tamao
void DestruirTHash (TablaHashT TablaHash, int N)
{
  int indice;
  for (indice = 0; indice < N; indice++)
    {
      if (TablaHash[indice].Palabra == NULL)
	{
          delete TablaHash[indice].Palabra;	//Destruir cada objeto activo de la tabla
	}//EndIf
    }//EndFor
  delete TablaHash;				//Destruir la tabla.
}//EndDestruirTHash


//Funcin crearclave: Crea una clave para una palabra dada
//Variables de entrada: la palabra
//Devuelve el valor de la clave para la palabra
unsigned long int crearclave (Elemento ElementoH)
{
  size_t largo;					//Largo de palabra
  unsigned int i;				//Contador
  unsigned long int clave = 0;			//Valor de la clave
  unsigned short int letra;			//Valor ASCII de una letra

  largo = strlen (ElementoH.Palabra);		//Primero, calculamos el largo de la palabra
  for (i = 0; i < largo; i++)			//Para cada letra de la palabra...
      {
	letra = ElementoH.Palabra[i];		//...obtenemos su valor ASCII...
        clave = clave + 32*letra;		//...lo multiplicamos por 32 y lo sumamos a la clave
        if (i < (largo - 1))			//Es la ltima letra de la palabra?
           {
             letra = ElementoH.Palabra[i+1];	//Si no, entonces obtenemos el valor ASCII de la sig. letra...
             clave = clave + letra;		//...y se lo sumamos a la clave
	   }//EndIf
      }//EndFor
  return (clave);
}//Endcrearclave
